import logging
from datetime import timedelta
from typing import Annotated

from fastapi import Depends, FastAPI, HTTPException, status, responses
from sqlalchemy.orm import selectinload
from fastapi.security import OAuth2PasswordRequestForm
from sqlmodel import select

from gisaf.models.authentication import (
  User, UserRead,
  Role, RoleRead,
)
from gisaf.models.category import Category, CategoryRead
from gisaf.models.to_migrate import DataProvider
from gisaf.config import conf
from gisaf.models.bootstrap import BootstrapData
from gisaf.models.store import Store
from gisaf.models.project import Project
from gisaf.database import pandas_query, fastapi_db_session as db_session
from gisaf.security import (
    Token,
    authenticate_user, get_current_user, create_access_token,
    )
from gisaf.registry import registry, NotInRegistry
from gisaf.redis_tools import store as redis_store
from gisaf.custom_store_base import BaseStore
from gisaf.models.to_migrate import (
  FeatureInfo, InfoItem, Attachment, InfoCategory
)

logger = logging.getLogger(__name__)


api = FastAPI(
    default_response_class=responses.ORJSONResponse,
)
#api.add_middleware(SessionMiddleware, secret_key=conf.crypto.secret)

#db_session = Annotated[AsyncSession, Depends(get_db_session)]


@api.get('/bootstrap')
async def bootstrap(
    user: Annotated[UserRead, Depends(get_current_user)]) -> BootstrapData:
    return BootstrapData(user=user)


@api.post("/token")
async def login_for_access_token(
    form_data: OAuth2PasswordRequestForm = Depends()
    ) -> Token:
    user = await authenticate_user(form_data.username, form_data.password)
    if not user:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Bearer"},
        )
    access_token = create_access_token(
        data={"sub": user.username},
        expires_delta=timedelta(seconds=conf.crypto.expire))
    return Token(access_token=access_token, token_type='bearer')

@api.get("/list")
async def list_data_providers() -> list[DataProvider]:
    """
    Return a list of data providers, for use with the api (graphs, etc)
    :return:
    """
    return [
        DataProvider(
            name=model.get_store_name(),
            values=[value.get_store_name() for value in values]
        ) for model, values in registry.values_for_model.items()]


@api.get("/users")
async def get_users(
    db_session: db_session,
    ) -> list[UserRead]:
    query = select(User).options(selectinload(User.roles))
    data = await db_session.exec(query)
    return data.all()

@api.get("/roles")
async def get_roles(
    db_session: db_session,
    ) -> list[RoleRead]:
    query = select(Role).options(selectinload(Role.users))
    data = await db_session.exec(query)
    return data.all()

@api.get("/categories")
async def get_categories(
    db_session: db_session,
    ) -> list[CategoryRead]:
    query = select(Category)
    data = await db_session.exec(query)
    return data.all()

@api.get("/categories_pandas")
async def get_categories_p(
    db_session: db_session,
    ) -> list[CategoryRead]:
    query = select(Category)
    df = await db_session.run_sync(pandas_query, query)
    return df.to_dict(orient="records")

@api.get("/stores")
async def get_stores() -> list[Store]:
    df = registry.stores.reset_index().\
        drop(columns=['model', 'raw_model', 'base_gis_type'])
    return df.to_dict(orient="records")

@api.get("/projects")
async def get_projects(
    db_session: db_session,
    ) -> list[Project]:
    query = select(Project)
    df = await db_session.run_sync(pandas_query, query)
    return df.to_dict(orient="records")

@api.get("/feature-info/{store}/{id}")
async def get_feature_info(
    store: str, id: str,
    ) -> FeatureInfo:
    store_record = registry.stores.loc[store]
    model = store_record.model

    if store_record.is_live:
        item = await redis_store.get_feature_info(store, id)
        geom = item.geometry
        ## Reproject to default coordinate system (WGS84),
        ## XXX: only for shapely geometries
        if not isinstance(geom, pygeos.Geometry):
            geom_reprojected = transform(reproject_func, geom)
        geoInfoItems = OrderedDict()
        if isinstance(geom, Point):
            geoInfoItems['longitude'] = f'{geom.x:.6f}'
            geoInfoItems['latitude'] = f'{geom.y:.6f}'
            if geom.has_z:
                geoInfoItems['elevation (m)'] = f'{geom.z:.6f}'
        elif isinstance(geom, (LineString, MultiLineString)):
            bounds = geom.bounds
            geoInfoItems['longitude'] = f'{bounds[0]:.6f} - {bounds[2]:.6f}'
            geoInfoItems['latitude'] = f'{bounds[1]:.6f} - {bounds[3]:.6f}'
            geoInfoItems['length (m)'] = f'{geom_reprojected.length:.2f}'
            ## TODO: elevation for MultiLineString
            if geom.has_z and not isinstance(geom, MultiLineString):
                elevations = [cc[2] for cc in geom.coords]
                elev_min = min(elevations)
                elev_max = max(elevations)
                if elev_min == elev_max:
                    geoInfoItems['elevation (m)'] = f'{elev_min:.2f}'
                else:
                    geoInfoItems['elevation (m)'] = f'{elev_min:.2f} - {elev_max:.2f}'
        elif isinstance(geom, (Polygon, MultiPolygon)):
            area = geom_reprojected.area
            bounds = geom.bounds
            geoInfoItems['longitude'] = f'{bounds[0]:.6f} - {bounds[2]:.6f}'
            geoInfoItems['latitude'] = f'{bounds[1]:.6f} - {bounds[3]:.6f}'
            geoInfoItems['area (sq. m)'] = f'{area:.1f} sq. m'
            geoInfoItems['area (ha)'] = f'{area / 10000:.1f} ha'
            geoInfoItems['area (acre)'] = f'{area / 4046.85643005078874:.1f} acres'
            ## TODO: elevation for MultiPolygon
            if geom.has_z and not isinstance(geom, MultiPolygon):
                if hasattr(geom, 'exterior'):
                    coords = geom.exterior.coords
                else:
                    coords = geom.coords
                elevations = [coord[2] for coord in coords]
                elev_min = min(elevations)
                elev_max = max(elevations)
                if elev_min == elev_max:
                    geoInfoItems['elevation (m)'] = f'{elev_min:.2f}'
                else:
                    geoInfoItems['elevation (m)'] = f'{elev_min:.2f} - {elev_max:.2f}'

        feature_info_dict = {
            'itemName': item.get('popup', f'Live: {store} #{id}'),
            'geoInfoItems': geoInfoItems,
            'surveyInfoItems': {
                'Note': 'Live layers do not have survey info',
            },
            'infoItems': dict(item.drop(set(item.keys()).intersection(('geometry', 'popup')))),
            'tags': {},
        }
    elif issubclass(model, BaseStore):
        feature_info_dict = await model.get_item_params(id)
    else:
        ## Not a live layer
        try:
            feature_info_dict = await registry.get_model_id_params(model, int(id))
        except NotInRegistry:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    feature_info = FeatureInfo(
        id=id,
        itemName=feature_info_dict['itemName'],
        geoInfoItems=[InfoItem(key=k, value=v)
                    for k, v in feature_info_dict['geoInfoItems'].items()],
        surveyInfoItems=[InfoItem(key=k, value=v)
                    for k, v in feature_info_dict['surveyInfoItems'].items()],
        infoItems=[InfoItem(key=k, value=v)
                    for k, v in feature_info_dict['infoItems'].items()],
        tags=[InfoItem(key=k, value=v)
                    for k, v in feature_info_dict['tags'].items()],
        graph=feature_info_dict.get('graph'),
    )
    if 'files' in feature_info_dict and feature_info_dict['files'] is not None:
        feature_info.files = [
            Attachment(name=k, path=v)
            for k, v in feature_info_dict['files'].items()
        ]
    else:
        feature_info.files = []
    if 'images' in feature_info_dict and feature_info_dict['images'] is not None:
        feature_info.images = [
            Attachment(name=k, path=v)
            for k, v in feature_info_dict['images'].items()
        ]
    else:
        feature_info.images = []
    if 'categorized_info_items' in feature_info_dict and feature_info_dict['categorized_info_items'] != None:
        feature_info.categorizedInfoItems = [
            InfoCategory(
                name=name,
                infoItems=[
                    InfoItem(key=k, value=v)
                    for k, v in info_items.items()
                ]
            )
            for name, info_items in feature_info_dict['categorized_info_items'].items()
        ]
    feature_info.externalRecordUrl = feature_info_dict.get('externalRecordUrl')
    return feature_info

# @api.get("/user-role")
# async def get_user_role_relation(
#     *, db_session: AsyncSession = Depends(get_db_session)
#     ) -> list[UserRoleLink]:
#     roles = await db_session.exec(select(UserRoleLink))
#     return roles.all()